#region MigraDoc - Creating Documents on the Fly
//
// Authors:
//   Stefan Lange
//   Klaus Potzesny
//   David Stephensen
//
// Copyright (c) 2001-2017 empira Software GmbH, Cologne Area (Germany)
//
// http://www.pdfsharp.com
// http://www.migradoc.com
// http://sourceforge.net/projects/pdfsharp
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
#endregion

using MigraDoc.DocumentObjectModel.publics;
using MigraDoc.DocumentObjectModel.Visitors;

namespace MigraDoc.DocumentObjectModel
{
    /// <summary>
    /// Represents the collection of HeaderFooter objects.
    /// </summary>
    public class HeadersFooters : DocumentObject, IVisitable
    {
        /// <summary>
        /// Initializes a new instance of the HeadersFooters class.
        /// </summary>
        public HeadersFooters()
        { }

        /// <summary>
        /// Initializes a new instance of the HeadersFooters class with the specified parent.
        /// </summary>
        public HeadersFooters(DocumentObject parent) : base(parent) { }

        #region Methods
        /// <summary>
        /// Creates a deep copy of this object.
        /// </summary>
        public new HeadersFooters Clone()
        {
            return (HeadersFooters)DeepCopy();
        }

        /// <summary>
        /// Implements the deep copy of the object.
        /// </summary>
        protected override object DeepCopy()
        {
            HeadersFooters headersFooters = (HeadersFooters)base.DeepCopy();
            if (headersFooters._evenPage != null)
            {
                headersFooters._evenPage = headersFooters._evenPage.Clone();
                headersFooters._evenPage._parent = headersFooters;
            }
            if (headersFooters._firstPage != null)
            {
                headersFooters._firstPage = headersFooters._firstPage.Clone();
                headersFooters._firstPage._parent = headersFooters;
            }
            if (headersFooters._primary != null)
            {
                headersFooters._primary = headersFooters._primary.Clone();
                headersFooters._primary._parent = headersFooters;
            }
            return headersFooters;
        }
        #endregion

        #region Properties
        /// <summary>
        /// Returns true if this collection contains headers, false otherwise.
        /// </summary>
        public bool IsHeader
        {
            get
            {
                Section sec = (Section)_parent;
                return sec._headers == this;
            }
        }

        /// <summary>
        /// Returns true if this collection contains footers, false otherwise.
        /// </summary>
        public bool IsFooter
        {
            get { return !IsHeader; }
        }

        /// <summary>
        /// Determines whether a particular header or footer exists.
        /// </summary>
        public bool HasHeaderFooter(HeaderFooterIndex index)
        {
            return !IsNull(index.ToString());
        }

        /// <summary>
        /// Gets or sets the even page HeaderFooter of the HeadersFooters object.
        /// </summary>
        public HeaderFooter EvenPage
        {
            get { return _evenPage ?? (_evenPage = new HeaderFooter(this)); }
            set
            {
                SetParent(value);
                _evenPage = value;
            }
        }
        [DV]
        public HeaderFooter _evenPage;

        /// <summary>
        /// Gets or sets the first page HeaderFooter of the HeadersFooters object.
        /// </summary>
        public HeaderFooter FirstPage
        {
            get { return _firstPage ?? (_firstPage = new HeaderFooter(this)); }
            set
            {
                SetParent(value);
                _firstPage = value;
            }
        }
        [DV]
        public HeaderFooter _firstPage;

        /// <summary>
        /// Gets or sets the primary HeaderFooter of the HeadersFooters object.
        /// </summary>
        public HeaderFooter Primary
        {
            get { return _primary ?? (_primary = new HeaderFooter(this)); }
            set
            {
                SetParent(value);
                _primary = value;
            }
        }
        [DV]
        public HeaderFooter _primary;
        #endregion

        #region public
        /// <summary>
        /// Converts HeadersFooters into DDL.
        /// </summary>
        public override void Serialize(Serializer serializer)
        {
            bool hasPrimary = HasHeaderFooter(HeaderFooterIndex.Primary);
            bool hasEvenPage = HasHeaderFooter(HeaderFooterIndex.EvenPage);
            bool hasFirstPage = HasHeaderFooter(HeaderFooterIndex.FirstPage);

            // \primary...
            if (hasPrimary)
                Primary.Serialize(serializer, "primary");

            // \even... 
            if (hasEvenPage)
                EvenPage.Serialize(serializer, "evenpage");

            // \firstpage...
            if (hasFirstPage)
                FirstPage.Serialize(serializer, "firstpage");
        }

        /// <summary>
        /// Allows the visitor object to visit the document object and its child objects.
        /// </summary>
        void IVisitable.AcceptVisitor(DocumentObjectVisitor visitor, bool visitChildren)
        {
            visitor.VisitHeadersFooters(this);

            if (visitChildren)
            {
                if (HasHeaderFooter(HeaderFooterIndex.Primary))
                    ((IVisitable)_primary).AcceptVisitor(visitor, true);
                if (HasHeaderFooter(HeaderFooterIndex.EvenPage))
                    ((IVisitable)_evenPage).AcceptVisitor(visitor, true);
                if (HasHeaderFooter(HeaderFooterIndex.FirstPage))
                    ((IVisitable)_firstPage).AcceptVisitor(visitor, true);
            }
        }

        /// <summary>
        /// Returns the meta object of this instance.
        /// </summary>
        public override Meta Meta
        {
            get { return _meta ?? (_meta = new Meta(typeof(HeadersFooters))); }
        }
        static Meta _meta;
        #endregion
    }
}
